#ifndef _3DOBJ_CODE
#define _3DOBJ_CODE

#include <stdio.h>    // NULL, fopen(), fclose(), fgetc()
#include <ctype.h>    // isspace(),
#include <windows.h>
#include "fix.h"      // Precalculated 3D data
#include "3dobj.h"    // 3D class definition

// ***************************************************************************
// POLY.CPP
//   A set of functions to create translation matrices for various
// polygon transformation functions, to transform a polygon-fill
// object using a matrix concatenated from those matrices, and to
// draw a polygon-fill object on the mode 13h display
// ***************************************************************************

// Used in calculating matrices, and fixed point calculations within
// matrices. SHIFT const defines the accuracy of the calculation.
#define   ONE    (1L<<SHIFT)

// ***************************************************************************
// ************************* 3D Object Member Functions***********************
// ***************************************************************************

// ***************************************************************************
// ************************* Initialisation Functions ************************
// ***************************************************************************

// Constructor
Obj3D::Obj3D ()
{
    Initialise ();
}

void Obj3D::Initialise (void)
{
    pObject3D = NULL;
    // Set update region to offscreen area.
    updateRegion.xMin = updateRegion.xMax =     
    updateRegion.yMin = updateRegion.yMax = -1;
}

// ***************************************************************************
// *************************** Load Data Functions ***************************
// ***************************************************************************
int Obj3D::Load3DObj (FILE* pFileName)
{
    // Remove current 3D object info
    FreeObject3DMem ();

    LoadPoly (pFileName);
    return 0;
}

int Obj3D::Load3DObj (ObjectType* pObj3D)
{
    // Remove current 3D object info
    FreeObject3DMem ();

    pObject3D = pObj3D;
    return 0;
}

// ***************************************************************************
// ********************** Set 3D Object View Parameters  *********************
// ***************************************************************************
// Transformation functions:
// Initialize master transformation matrix to the identity matrix
void Obj3D::InitTransformation()
{
    matrix[0][0] = ONE; matrix[0][1] = 0;   matrix[0][2] = 0;   matrix[0][3] = 0;
    matrix[1][0] = 0;   matrix[1][1] = ONE; matrix[1][2] = 0;   matrix[1][3] = 0;
    matrix[2][0] = 0;   matrix[2][1] = 0;   matrix[2][2] = ONE; matrix[2][3] = 0;
    matrix[3][0] = 0;   matrix[3][1] = 0;   matrix[3][2] = 0;   matrix[3][3] = ONE;
}

// Set Scale factor.
void Obj3D::SetScaleMatrix (long sf)
{
    long mat[4][4];

    // Initialize scaling matrix:
    smat[0][0] = sf << SHIFT;   smat[0][1] = 0;           smat[0][2] = 0;           smat[0][3] = 0;
    smat[1][0] = 0;             smat[1][1] = sf << SHIFT; smat[1][2] = 0;           smat[1][3] = 0;
    smat[2][0] = 0;             smat[2][1] = 0;           smat[2][2] = sf << SHIFT; smat[2][3] = 0;
    smat[3][0] = 0;             smat[3][1] = 0;           smat[3][2] = 0;           smat[3][3] = ONE;

    // Concatenate with master matrix:
    MatMult (mat, smat, matrix);
    MatCopy (matrix, mat);
}

// Set rotation factors.
void Obj3D::SetRotationMatrix   (int ax, int ay, int az)
{
    // Create three rotation matrices that will rotate an object
    // AX radians on the X axis, AY radians on the Y axis and
    // AZ radians on the Z axis

    long mat1[4][4];
    long mat2[4][4];

    // Initialize X rotation matrix:
    xmat[0][0] = ONE;  xmat[0][1] = 0;        xmat[0][2] = 0;       xmat[0][3] = 0;
    xmat[1][0] = 0;    xmat[1][1] = COS(ax);  xmat[1][2] = SIN(ax); xmat[1][3] = 0;
    xmat[2][0] = 0;    xmat[2][1] = -SIN(ax); xmat[2][2] = COS(ax); xmat[2][3] = 0;
    xmat[3][0] = 0;    xmat[3][1] = 0;        xmat[3][2] = 0;       xmat[3][3] = ONE;

    // Concatenate this matrix with master matrix:
    MatMult(mat1, xmat, matrix);

    // Initialize Y rotation matrix:
    ymat[0][0] = COS(ay); ymat[0][1] = 0;   ymat[0][2] = -SIN(ay); ymat[0][3] = 0;
    ymat[1][0] = 0;       ymat[1][1] = ONE; ymat[1][2] = 0;        ymat[1][3] = 0;
    ymat[2][0] = SIN(ay); ymat[2][1] = 0;   ymat[2][2] = COS(ay);  ymat[2][3] = 0;
    ymat[3][0] = 0;       ymat[3][1] = 0;   ymat[3][2] = 0;        ymat[3][3] = ONE;

    // Concatenate this matrix with master matrix:
    MatMult(mat2, ymat, mat1);

    // Initialize Z rotation matrix:
    zmat[0][0] = COS(az);  zmat[0][1] = SIN(az);  zmat[0][2] = 0;    zmat[0][3] = 0;
    zmat[1][0] = -SIN(az); zmat[1][1] = COS(az);  zmat[1][2] = 0;    zmat[1][3] = 0;
    zmat[2][0] = 0;        zmat[2][1] = 0;        zmat[2][2] = ONE;  zmat[2][3] = 0;
    zmat[3][0] = 0;        zmat[3][1] = 0;        zmat[3][2] = 0;    zmat[3][3] = ONE;

    // Concatenate this matrix with master matrix:
    MatMult(matrix, zmat, mat2);
}

// Move object around 3D world
void Obj3D::SetTranslateMatrix  (long xt, long yt, long zt)
{
    // Create a translation matrix that will translate an object an
    // X distance of XT, a Y distance of YT, and a Z distance of ZT
    // from the screen origin
    long mat[4][4];

    tmat[0][0] = ONE;         tmat[0][1] = 0;           tmat[0][2] = 0;           tmat[0][3] = 0;
    tmat[1][0] = 0;           tmat[1][1] = ONE;         tmat[1][2] = 0;           tmat[1][3] = 0;
    tmat[2][0] = 0;           tmat[2][1] = 0;           tmat[2][2] = ONE;         tmat[2][3] = 0;
    tmat[3][0] = xt << SHIFT; tmat[3][1] = yt << SHIFT; tmat[3][2] = zt << SHIFT; tmat[3][3] = ONE;

    // Concatenate with master matrix:
    MatMult(mat, matrix, tmat);
    MatCopy(matrix, mat);
}

// Transform and project object with matrices.
void Obj3D::Transform           (void)
{
    // Multiply all vertices in OBJECT with master transformation matrix:
    for (int v = 0; v < pObject3D -> numberOfVertices; v++)
    {
        VertexType *vptr = (pObject3D -> vertex) + v;

        vptr->wx = (vptr->lx * matrix[0][0] + vptr->ly * matrix[1][0] + vptr->lz * matrix[2][0] + matrix[3][0]) >> SHIFT;
        vptr->wy = (vptr->lx * matrix[0][1] + vptr->ly * matrix[1][1] + vptr->lz * matrix[2][1] + matrix[3][1]) >> SHIFT;
        vptr->wz = (vptr->lx * matrix[0][2] + vptr->ly * matrix[1][2] + vptr->lz * matrix[2][2] + matrix[3][2]) >> SHIFT;
    }
}

void Obj3D::SetPerspectProj     (long distance, long xOrigin, long yOrigin)
{
    // Project object onto screen with perspective projection

    // Loop through vertices
    for (int v = 0; v <  pObject3D -> numberOfVertices; v++)
    {
        VertexType *vptr = (pObject3D -> vertex) + v;   // Point to current vertex

        // ...divide world x&y coords by z coordinates
        vptr->sx = distance * vptr->wx / (distance - vptr->wz) + xOrigin;
        vptr->sy = distance * vptr->wy / (distance - vptr->wz) + yOrigin;
    }
}

void Obj3D::DrawObject (HDC hdc)
{
    // Draw object in structure OBJECT
    // reset region structure items.
    updateRegion.xMin = updateRegion.yMin = 0x7fff;    // 32000 approx
    updateRegion.xMax = updateRegion.yMax = -1;

    // Loop through all polygons in object:
    for (int p=0; p < pObject3D -> numberOfPolygons; p++)
    {
        // Draw current polygon:
        if (pObject3D -> convex)
        {
            if(!BackFace(pObject3D -> polygon[p]))
                DrawPoly(pObject3D -> polygon[p], pObject3D -> convex, hdc);
        }
        else
            DrawPoly(pObject3D -> polygon[p], pObject3D -> convex, hdc);
    }
}

// ***************************************************************************
// ***************** User 3D Display Interface Function **********************
// ***************************************************************************
void Obj3D::Show (Display3DParams& params, HDC hdc)
{
    InitTransformation();                                // Init master matrix.
    SetScaleMatrix    (params.sf);                       // Create scaling matrix
    SetRotationMatrix (params.rx, params.ry, params.rz); // Create rotation matrix
    SetTranslateMatrix(params.tx, params.ty, params.tz); // Translate object:
    Transform();                                         // Transform object with matrix:
    SetPerspectProj   (params.zDist,
                       params.xOrigin, params.yOrigin);  // Do perspective projection:

    // Draw the object:
    DrawObject(hdc);
}

// ***************************************************************************
// *************************** Destructor ************************************
// ***************************************************************************
// Clear up memory.
void Obj3D::FreeObject3DMem (void)
{
    if (pObject3D != NULL)
    {
        // Delete polygons
        for (int delPoly = 0; delPoly < pObject3D -> numberOfPolygons; delPoly++)
        {
            // Delete vertice INDEX POINTER array within polygon
            // Delete array, not arrays pointers to vertices, these
            // will be delete later.
            delete pObject3D -> polygon[delPoly].vertex;
        }

        // Delete polygon array.
        delete pObject3D -> polygon; // Think this should be delete pObject3D -> polgon[]

        // Delete vertices array list
        delete pObject3D -> vertex;
    }
}

// Destructor
Obj3D::~Obj3D ()
{
    FreeObject3DMem ();
}

// ***************************************************************************
// *********** General Matrix Functions Used Within This File ****************
// ***************************************************************************
void Obj3D::MatMult(long result[4][4],long mat1[4][4],long mat2[4][4])
{
    // Multiply matrix MAT1 by matrix MAT2,
    //  returning the result in RESULT

    for (int i=0; i<4; i++)
    {
        for (int j=0; j<4; j++)
        {
            result[i][j] = ( (mat1[i][0] * mat2[0][j]) +
                             (mat1[i][1] * mat2[1][j]) +
                             (mat1[i][2] * mat2[2][j]) +
                             (mat1[i][3] * mat2[3][j]) ) >> SHIFT;
        }
    }
}


void Obj3D::MatCopy(long dest[4][4],long source[4][4])
{
    // Copy matrix SOURCE to matrix DEST
    for (int x = 0; x < 4; x++)
    {
        for (int y = 0; y < 4; y++)
                dest[x][y] = source[x][y];
    }
}

// ***************************************************************************
// ************* Draw Polygon Functions *****************************************
// ***************************************************************************
int Obj3D::BackFace(PolygonType& p)
{
    //   Returns 0 if POLYGON is visible, -1 if not.
    //   POLYGON must be part of a convex polyhedron

    VertexType *v0,*v1,*v2;                // Pointers to three vertices
    // Point to vertices:
    v0 = p.vertex[0];
    v1 = p.vertex[1];
    v2 = p.vertex[2];

    long z = ((v1->sx - v0->sx) * (v2->sy  - v0->sy)) -
             ((v1->sy - v0->sy) * (v2->sx  - v0->sx));
    return(z >= 0);
}

RectUD* Obj3D::GetUpdateRegion (void)
{
    return &updateRegion;
}

void Obj3D::DrawPoly (PolygonType& polygon, int convex, HDC hdc)
{
    POINT* pPoints = new POINT[polygon.numberOfVertices];

    if (pPoints == NULL)
       return;

    // copy over polgon points into windows polydraw function.
    for (int vNum = 0; vNum < polygon.numberOfVertices; vNum++)
    {
        pPoints[vNum].x = (int)polygon.vertex[vNum] -> sx;
        pPoints[vNum].y = (int)polygon.vertex[vNum] -> sy;

        // check cordinates to find bounding rectangle for updating
        if (pPoints[vNum].x < updateRegion.xMin)
           updateRegion.xMin = pPoints[vNum].x;

        if (pPoints[vNum].x > updateRegion.xMax)
           updateRegion.xMax = pPoints[vNum].x;

        if (pPoints[vNum].y < updateRegion.yMin)
           updateRegion.yMin = pPoints[vNum].y;

        if (pPoints[vNum].y > updateRegion.yMax)
           updateRegion.yMax = pPoints[vNum].y;
    }

    // Create pen, and brush and sert to polygon colour.
    HBRUSH hOldBrush;
    HBRUSH hbr;
    HPEN   hOldPen;
    HPEN   hpen;
    
    if (convex == 0)    // Display as wire frame polygons
        hbr = GetStockObject  (NULL_BRUSH);
    else
        hbr = CreateSolidBrush(PALETTEINDEX(polygon.color));

    hpen = CreatePen       (PS_SOLID, 1, RGB (255, 255, 255));

    // Select objects to be used.
    hOldBrush = SelectObject (hdc, hbr);
    hOldPen   = SelectObject (hdc, hpen);

    // draw polygon on screen.
    Polygon(hdc, pPoints, polygon.numberOfVertices);

    // Clean up GDI objects !
    DeleteObject (SelectObject (hdc, hOldBrush));
    DeleteObject (SelectObject (hdc, hOldPen));

    delete pPoints;
}

// ***************************************************************************
// ************* 3D Object Load Functions ************************************
// ***************************************************************************
char Obj3D::NextChar(FILE *f)
{
    char ch;

    while(!feof(f))
    {
        while(isspace(ch = fgetc(f)));

        if (ch == '*')
            while((ch = fgetc(f)) != '\n');
        else
            return(ch);
    }
    return(0);
}

int Obj3D::GetNumber(FILE *f)
{
    char ch;
    int sign=1;
    int num = 0;

    if ((ch = NextChar(f)) == '-')
    {
        sign = -1;
        ch = NextChar(f);
    }
    while (isdigit(ch))
    {
        num = num*10+ch-'0';
        ch  = NextChar(f);
    }
    return(num*sign);
}

int Obj3D::LoadPoly (FILE* f)
{
    // Create new object structure.
    if (pObject3D != NULL)
       FreeObject3DMem ();

    pObject3D = new ObjectType;

    // Obtain vertice list for object
    pObject3D -> numberOfVertices = GetNumber(f);

    // Allocate required memory.
    pObject3D -> vertex = new VertexType[pObject3D -> numberOfVertices];

    // Load vertice information from file.
    for (int vertnum = 0; vertnum < pObject3D -> numberOfVertices; vertnum++)
    {
            pObject3D -> vertex[vertnum].lx = (long)GetNumber(f);
            pObject3D -> vertex[vertnum].ly = (long)GetNumber(f);
            pObject3D -> vertex[vertnum].lz = (long)GetNumber(f);
            pObject3D -> vertex[vertnum].lt = 1L;
            pObject3D -> vertex[vertnum].wt = 1L;
    }

    // Load polygon information.
    pObject3D -> numberOfPolygons = GetNumber(f);
    pObject3D -> polygon = new PolygonType[pObject3D -> numberOfPolygons];

    for (int polynum = 0; polynum < pObject3D -> numberOfPolygons; polynum++)
    {
        // Load vertice pointer index information for polygon.
        pObject3D -> polygon[polynum].numberOfVertices   = GetNumber(f);
        pObject3D -> polygon[polynum].vertex             = new VertexType* [pObject3D -> polygon[polynum].numberOfVertices];
        for (int vertnum = 0; vertnum < pObject3D -> polygon[polynum].numberOfVertices; vertnum++)
        {
            // Assign a POINTER INDEX for polygon to object list of vertices (vertex structures).
            pObject3D -> polygon[polynum].vertex[vertnum] = &(pObject3D -> vertex[GetNumber(f)]);
        }
        pObject3D -> polygon[polynum].color = GetNumber(f);
    }

    // Is 3D object convex.
    pObject3D -> convex = GetNumber(f);

    return(0);
}


#endif // _3DOBJ_CODE