#ifndef _3DOBJ_HEADER
#define _3DOBJ_HEADER

#include <stdio.h>
#include <windows.h>

// Constant definitions:
#define PARALLEL     0                           // Orthographic projection flag
#define PERSPECTIVE  1                           // Perspective projection flag

// Variable structures to hold shape data:
struct VertexType                                // Structure for individual vertices
{
    long lx, ly, lz, lt;                         // Local coordinates of vertex
    long wx, wy, wz, wt;                         // World coordinates of vertex
    long sx, sy, st;                             // Screen coordinates of vertex
};

struct PolygonType
{
    int           numberOfVertices;              // Number of vertices in polygon
    int           color;                         // Color of polygon
    long          zMax, zMin;                    // Maximum and minimum z coordinates of polygon
    long          xMax, xMin;
    long          yMax, yMin;
    VertexType**  vertex;                        // List of vertices (index pointers to polgon vertice array)
    int           sortFlag;                      // For hidden surface sorts
};

// Structure used in displaying a 3D object
struct Display3DParams
{
    long sf;                     // Scale factor
    int  rx,    ry,      rz;     // rotation of object in space
    long tx,    ty,      tz;     // Transation of object in space from origin
    long zDist, xOrigin, yOrigin;// Perspective viewing of object.
};

// *********************** Update Structure ***************************
// Structure used to contain the bounding rectangle of an object. This
// is used when updating virtual and real screens.
struct RectUD   // Rectangle update date region.
{
    int xMin, xMax;
    int yMin, yMax;
}; 

// Stores actual data of 3D object.
struct ObjectType
{
    int           numberOfVertices;              // Number of vertices in object
    int           numberOfPolygons;              // Number of polygons in object
    long          x, y, z;                       // World coordinates of object's local origin
    PolygonType*  polygon;                       // List of polygons in object
    VertexType*   vertex;                        // Array of vertices in object
    int           convex;                        // Is it a convex polyhedron?
};

class Obj3D
{
    protected:
        // Actual 3D data.
        ObjectType*  pObject3D;

        // Global transformation arrays:
        long matrix[4][4];                                // Master transformation matrix
        long smat[4][4];                                  // Scaling matrix
        long zmat[4][4];                                  // Z rotation matrix
        long xmat[4][4];                                  // X rotation matrix
        long ymat[4][4];                                  // Y rotation matrix
        long tmat[4][4];                                  // Translation matrix

        // Update region variables for displaying/erasing objects fast.
        RectUD    updateRegion;

        // **** Internal functions ****
        // General Matrix calculation functions.
        void MatMult  (long result[4][4],long mat1[4][4], long mat2[4][4]);
        void MatCopy  (long dest[4][4],  long source[4][4]);

        // Polygon removal and display functions.
        int  BackFace (PolygonType& p);
        void DrawPoly (PolygonType& polygon, int convex, HDC hdc);

        // Load 3D object from file.
        char NextChar (FILE *f);
        int  GetNumber(FILE *f);
        int  LoadPoly (FILE* f);

        // ********************** Transformation functions *******************
        // Initialize master transformation matrix to the identity matrix
        void InitTransformation  (void);
        void SetScaleMatrix      (long sf);                   // Set Scale factor.
        void SetRotationMatrix   (int  ax, int  ay, int  az); // Set rotation factors.
        void SetTranslateMatrix  (long xt, long yt, long zt); // Move object around 3D world.
        void Transform           (void);
        void SetPerspectProj     (long distance, long xOrigin, long yOrigin);
        void DrawObject          (HDC hdc);
        // *******************************************************************


        // Free 3D object function.
        void FreeObject3DMem (void);

    public:
        // Constructor
        Obj3D();
        void Initialise (void);

        // *******************************************************************
        // Load 3D object.
        int Load3DObj  (FILE* pFileName);
        int Load3DObj  (ObjectType* pObj3D);
        // *******************************************************************

        // *******************************************************************
        // Display object.
        void Show (Display3DParams& params, HDC hdc);
        // *******************************************************************
        // Get region information.
        RectUD*  GetUpdateRegion (void);

        // *******************************************************************
        // Destructor
        virtual ~Obj3D();
};

#endif // _3DOBJ_HEADER